@@ -25,19 +25,21 @@
 // SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 use std::fmt::Display;
+
 use std::ops::Index;
 use std::ops::IndexMut;
 use std::ops::RangeInclusive;
-use std::time;
 
-use ring::aead;
+use std::time::Instant;
 
 use crate::Error;
 use crate::Result;
+use crate::DEFAULT_INITIAL_CONGESTION_WINDOW_PACKETS;
 
 use crate::crypto;
 use crate::rand;
 use crate::ranges;
+use crate::recovery;
 use crate::stream;
 
 const FORM_BIT: u8 = 0x80;
@@ -53,6 +55,13 @@ pub const MAX_PKT_NUM_LEN: usize = 4;
 
 const SAMPLE_LEN: usize = 16;
 
+// Set the min skip skip interval to 2x the default number of initial packet
+// count.
+const MIN_SKIP_COUNTER_VALUE: u64 =
+    DEFAULT_INITIAL_CONGESTION_WINDOW_PACKETS as u64 * 2;
+
+const RETRY_AEAD_ALG: crypto::Algorithm = crypto::Algorithm::AES128_GCM;
+
 #[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Debug)]
 pub enum Epoch {
     Initial     = 0,
@@ -201,23 +210,23 @@ impl<'a> ConnectionId<'a> {
     }
 }
 
-impl<'a> Default for ConnectionId<'a> {
+impl Default for ConnectionId<'_> {
     #[inline]
     fn default() -> Self {
         Self::from_vec(Vec::new())
     }
 }
 
-impl<'a> From<Vec<u8>> for ConnectionId<'a> {
+impl From<Vec<u8>> for ConnectionId<'_> {
     #[inline]
     fn from(v: Vec<u8>) -> Self {
         Self::from_vec(v)
     }
 }
 
-impl<'a> From<ConnectionId<'a>> for Vec<u8> {
+impl From<ConnectionId<'_>> for Vec<u8> {
     #[inline]
-    fn from(id: ConnectionId<'a>) -> Self {
+    fn from(id: ConnectionId<'_>) -> Self {
         match id.0 {
             ConnectionIdInner::Vec(cid) => cid,
             ConnectionIdInner::Ref(cid) => cid.to_vec(),
@@ -225,16 +234,16 @@ impl<'a> From<ConnectionId<'a>> for Vec<u8> {
     }
 }
 
-impl<'a> PartialEq for ConnectionId<'a> {
+impl PartialEq for ConnectionId<'_> {
     #[inline]
     fn eq(&self, other: &Self) -> bool {
         self.as_ref() == other.as_ref()
     }
 }
 
-impl<'a> Eq for ConnectionId<'a> {}
+impl Eq for ConnectionId<'_> {}
 
-impl<'a> AsRef<[u8]> for ConnectionId<'a> {
+impl AsRef<[u8]> for ConnectionId<'_> {
     #[inline]
     fn as_ref(&self) -> &[u8] {
         match &self.0 {
@@ -244,14 +253,14 @@ impl<'a> AsRef<[u8]> for ConnectionId<'a> {
     }
 }
 
-impl<'a> std::hash::Hash for ConnectionId<'a> {
+impl std::hash::Hash for ConnectionId<'_> {
     #[inline]
     fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
         self.as_ref().hash(state);
     }
 }
 
-impl<'a> std::ops::Deref for ConnectionId<'a> {
+impl std::ops::Deref for ConnectionId<'_> {
     type Target = [u8];
 
     #[inline]
@@ -263,14 +272,14 @@ impl<'a> std::ops::Deref for ConnectionId<'a> {
     }
 }
 
-impl<'a> Clone for ConnectionId<'a> {
+impl Clone for ConnectionId<'_> {
     #[inline]
     fn clone(&self) -> Self {
         Self::from_vec(self.as_ref().to_vec())
     }
 }
 
-impl<'a> std::fmt::Debug for ConnectionId<'a> {
+impl std::fmt::Debug for ConnectionId<'_> {
     #[inline]
     fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         for c in self.as_ref() {
@@ -403,12 +412,14 @@ impl<'a> Header<'a> {
             },
 
             Type::Retry => {
+                const TAG_LEN: usize = RETRY_AEAD_ALG.tag_len();
+
                 // Exclude the integrity tag from the token.
-                if b.cap() < aead::AES_128_GCM.tag_len() {
+                if b.cap() < TAG_LEN {
                     return Err(Error::InvalidPacket);
                 }
 
-                let token_len = b.cap() - aead::AES_128_GCM.tag_len();
+                let token_len = b.cap() - TAG_LEN;
                 token = Some(b.get_bytes(token_len)?.to_vec());
             },
 
@@ -522,7 +533,7 @@ impl<'a> Header<'a> {
     }
 }
 
-impl<'a> std::fmt::Debug for Header<'a> {
+impl std::fmt::Debug for Header<'_> {
     fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(f, "{:?}", self.ty)?;
 
@@ -555,20 +566,12 @@ impl<'a> std::fmt::Debug for Header<'a> {
     }
 }
 
-pub fn pkt_num_len(pn: u64) -> Result<usize> {
-    let len = if pn < u64::from(u8::MAX) {
-        1
-    } else if pn < u64::from(u16::MAX) {
-        2
-    } else if pn < 16_777_215u64 {
-        3
-    } else if pn < u64::from(u32::MAX) {
-        4
-    } else {
-        return Err(Error::InvalidPacket);
-    };
-
-    Ok(len)
+pub fn pkt_num_len(pn: u64, largest_acked: u64) -> usize {
+    let num_unacked: u64 = pn.saturating_sub(largest_acked) + 1;
+    // computes ceil of num_unacked.log2()
+    let min_bits = u64::BITS - num_unacked.leading_zeros();
+    // get the num len in bytes
+    min_bits.div_ceil(8) as usize
 }
 
 pub fn decrypt_hdr(
@@ -713,10 +716,10 @@ pub fn encrypt_pkt(
     Ok(payload_offset + ciphertext_len)
 }
 
-pub fn encode_pkt_num(pn: u64, b: &mut octets::OctetsMut) -> Result<()> {
-    let len = pkt_num_len(pn)?;
-
-    match len {
+pub fn encode_pkt_num(
+    pn: u64, pn_len: usize, b: &mut octets::OctetsMut,
+) -> Result<()> {
+    match pn_len {
         1 => b.put_u8(pn as u8)?,
 
         2 => b.put_u16(pn as u16)?,
@@ -784,26 +787,25 @@ pub fn retry(
 pub fn verify_retry_integrity(
     b: &octets::OctetsMut, odcid: &[u8], version: u32,
 ) -> Result<()> {
+    const TAG_LEN: usize = RETRY_AEAD_ALG.tag_len();
+
     let tag = compute_retry_integrity_tag(b, odcid, version)?;
 
-    ring::constant_time::verify_slices_are_equal(
-        &b.as_ref()[..aead::AES_128_GCM.tag_len()],
-        tag.as_ref(),
-    )
-    .map_err(|_| Error::CryptoFail)?;
-
-    Ok(())
+    crypto::verify_slices_are_equal(&b.as_ref()[..TAG_LEN], tag.as_ref())
 }
 
 fn compute_retry_integrity_tag(
     b: &octets::OctetsMut, odcid: &[u8], version: u32,
-) -> Result<aead::Tag> {
-    const RETRY_INTEGRITY_KEY_V1: [u8; 16] = [
+) -> Result<Vec<u8>> {
+    const KEY_LEN: usize = RETRY_AEAD_ALG.key_len();
+    const TAG_LEN: usize = RETRY_AEAD_ALG.tag_len();
+
+    const RETRY_INTEGRITY_KEY_V1: [u8; KEY_LEN] = [
         0xbe, 0x0c, 0x69, 0x0b, 0x9f, 0x66, 0x57, 0x5a, 0x1d, 0x76, 0x6b, 0x54,
         0xe3, 0x68, 0xc8, 0x4e,
     ];
 
-    const RETRY_INTEGRITY_NONCE_V1: [u8; aead::NONCE_LEN] = [
+    const RETRY_INTEGRITY_NONCE_V1: [u8; crypto::MAX_NONCE_LEN] = [
         0x46, 0x15, 0x99, 0xd3, 0x5d, 0x63, 0x2b, 0xf2, 0x23, 0x98, 0x25, 0xbb,
     ];
 
@@ -824,17 +826,23 @@ fn compute_retry_integrity_tag(
     pb.put_bytes(odcid)?;
     pb.put_bytes(&b.buf()[..hdr_len])?;
 
-    let key = aead::LessSafeKey::new(
-        aead::UnboundKey::new(&aead::AES_128_GCM, key)
-            .map_err(|_| Error::CryptoFail)?,
-    );
+    let key = crypto::PacketKey::new(
+        RETRY_AEAD_ALG,
+        key.to_vec(),
+        nonce.to_vec(),
+        crypto::Seal::ENCRYPT,
+    )?;
 
-    let nonce = aead::Nonce::assume_unique_for_key(nonce);
+    let mut out_tag = vec![0_u8; TAG_LEN];
 
-    let aad = aead::Aad::from(&pseudo);
+    let out_len = key.seal_with_u64_counter(0, &pseudo, &mut out_tag, 0, None)?;
 
-    key.seal_in_place_separate_tag(nonce, aad, &mut [])
-        .map_err(|_| Error::CryptoFail)
+    // Ensure that the output only contains the AEAD tag.
+    if out_len != out_tag.len() {
+        return Err(Error::CryptoFail);
+    }
+
+    Ok(out_tag)
 }
 
 pub struct KeyUpdate {
@@ -850,71 +858,71 @@ pub struct KeyUpdate {
     pub update_acked: bool,
 
     /// When the old key should be discarded.
-    pub timer: time::Instant,
+    pub timer: Instant,
 }
 
 pub struct PktNumSpace {
+    /// The largest packet number received.
     pub largest_rx_pkt_num: u64,
 
-    pub largest_rx_pkt_time: time::Instant,
+    /// Time the largest packet number received.
+    pub largest_rx_pkt_time: Instant,
 
+    /// The largest non-probing packet number.
     pub largest_rx_non_probing_pkt_num: u64,
 
-    pub next_pkt_num: u64,
+    /// The largest packet number send in the packet number space so far.
+    pub largest_tx_pkt_num: Option<u64>,
 
+    /// Range of packet numbers that we need to send an ACK for.
     pub recv_pkt_need_ack: ranges::RangeSet,
 
+    /// Tracks received packet numbers.
     pub recv_pkt_num: PktNumWindow,
 
+    /// Track if a received packet is ack eliciting.
     pub ack_elicited: bool,
-
-    pub key_update: Option<KeyUpdate>,
-
-    pub crypto_open: Option<crypto::Open>,
-    pub crypto_seal: Option<crypto::Seal>,
-
-    pub crypto_0rtt_open: Option<crypto::Open>,
-
-    pub crypto_stream: stream::Stream,
 }
 
 impl PktNumSpace {
     pub fn new() -> PktNumSpace {
         PktNumSpace {
             largest_rx_pkt_num: 0,
-
-            largest_rx_pkt_time: time::Instant::now(),
-
+            largest_rx_pkt_time: Instant::now(),
             largest_rx_non_probing_pkt_num: 0,
-
-            next_pkt_num: 0,
-
+            largest_tx_pkt_num: None,
             recv_pkt_need_ack: ranges::RangeSet::new(crate::MAX_ACK_RANGES),
-
             recv_pkt_num: PktNumWindow::default(),
-
             ack_elicited: false,
-
-            key_update: None,
-
-            crypto_open: None,
-            crypto_seal: None,
-
-            crypto_0rtt_open: None,
-
-            crypto_stream: stream::Stream::new(
-                0, // dummy
-                u64::MAX,
-                u64::MAX,
-                true,
-                true,
-                stream::MAX_STREAM_WINDOW,
-            ),
         }
     }
 
     pub fn clear(&mut self) {
-        self.crypto_stream = stream::Stream::new(
+        self.ack_elicited = false;
+    }
+
+    pub fn ready(&self) -> bool {
+        self.ack_elicited
+    }
+
+    pub fn on_packet_sent(&mut self, sent_pkt: &recovery::Sent) {
+        // Track the largest packet number sent
+        self.largest_tx_pkt_num =
+            self.largest_tx_pkt_num.max(Some(sent_pkt.pkt_num));
+    }
+}
+
+pub struct CryptoContext {
+    pub key_update: Option<KeyUpdate>,
+    pub crypto_open: Option<crypto::Open>,
+    pub crypto_seal: Option<crypto::Seal>,
+    pub crypto_0rtt_open: Option<crypto::Open>,
+    pub crypto_stream: stream::Stream,
+}
+
+impl CryptoContext {
+    pub fn new() -> CryptoContext {
+        let crypto_stream = stream::Stream::new(
             0, // dummy
             u64::MAX,
             u64::MAX,
@@ -922,23 +930,197 @@ impl PktNumSpace {
             true,
             stream::MAX_STREAM_WINDOW,
         );
+        CryptoContext {
+            key_update: None,
+            crypto_open: None,
+            crypto_seal: None,
+            crypto_0rtt_open: None,
+            crypto_stream,
+        }
+    }
 
-        self.ack_elicited = false;
+    pub fn clear(&mut self) {
+        self.crypto_open = None;
+        self.crypto_seal = None;
+        self.crypto_stream = <stream::Stream>::new(
+            0, // dummy
+            u64::MAX,
+            u64::MAX,
+            true,
+            true,
+            stream::MAX_STREAM_WINDOW,
+        );
+    }
+
+    pub fn data_available(&self) -> bool {
+        self.crypto_stream.is_flushable()
     }
 
     pub fn crypto_overhead(&self) -> Option<usize> {
         Some(self.crypto_seal.as_ref()?.alg().tag_len())
     }
 
-    pub fn ready(&self) -> bool {
-        self.crypto_stream.is_flushable() || self.ack_elicited
-    }
-
     pub fn has_keys(&self) -> bool {
         self.crypto_open.is_some() && self.crypto_seal.is_some()
     }
 }
 
+/// QUIC recommends skipping packet numbers to elicit a [faster ACK] or to
+/// mitigate an [optimistic ACK attack] (OACK attack). quiche currently skips
+/// packets only for the purposes of optimistic attack mitigation.
+///
+/// ## What is an Optimistic ACK attack
+/// A typical endpoint is responsible for making concurrent progress on multiple
+/// connections and needs to fairly allocate resources across those connection.
+/// In order to ensure fairness, an endpoint relies on "recovery signals" to
+/// determine the optimal sending rate per connection. For example, when a new
+/// flow joins a shared network, it might induce packet loss for other flows and
+/// cause those flows to yield bandwidth on the network. ACKs are the primary
+/// source of recovery signals for a QUIC connection.
+///
+/// The goal of an OACK attack is to mount a DDoS attack by exploiting recovery
+/// signals and causing a server to expand its sending rate. A server with an
+/// inflated sending rate would then be able to flood a shared network and
+/// cripple all other flows. The fundamental reason that makes OACK
+/// attach possible is the use of unvalidated ACK data, which is then used to
+/// modify internal state. Therefore at a high level, a mitigation should
+/// validate the incoming ACK data before use.
+///
+/// ## Optimistic ACK attack mitigation
+/// quiche follows the RFC's recommendation for mitigating an [optimistic ACK
+/// attack] by skipping packets and validating that the peer does NOT send ACKs
+/// for those skipped packets. If an ACK for a skipped packet is received, the
+/// connection is closed with a [PROTOCOL_VIOLATION] error.
+///
+/// A robust mitigation should skip packets randomly to ensure that an attacker
+/// can't predict which packet number was skipped. Skip/validation should happen
+/// "periodically" over the lifetime of the connection. Since skipping packets
+/// also elicits a faster ACK, we need to balance the skip frequency to
+/// sufficiently validate the peer without impacting other aspects of recovery.
+///
+/// A naive approach could be to skip a random packet number in the range
+/// 200-500 (pick some static range). While this might work, its not apparent if
+/// a static range is effective for all networks with varying bandwidths/RTTs.
+///
+/// Since an attacker can potentially influence the sending rate once per
+/// "round", it would be ideal to validate the peer once per round. Therefore,
+/// an ideal range seems to be one that dynamically adjusts based on packets
+/// sent per round, ie. adjust skip range based on the current CWND.
+///
+/// [faster ACK]: https://www.rfc-editor.org/rfc/rfc9002.html#section-6.2.4
+/// [optimistic ACK attack]: https://www.rfc-editor.org/rfc/rfc9000.html#section-21.4
+/// [PROTOCOL_VIOLATION]: https://www.rfc-editor.org/rfc/rfc9000#section-13.1
+pub struct PktNumManager {
+    // TODO:
+    // Defer including next_pkt_num in order to reduce the size of this patch
+    // /// Next packet number.
+    // next_pkt_num: u64,
+    /// Track if we have skipped a packet number.
+    skip_pn: Option<u64>,
+
+    /// Track when to skip the next packet number
+    ///
+    /// None indicates the counter is not armed while Some(0) indicates that the
+    /// counter has expired.
+    pub skip_pn_counter: Option<u64>,
+}
+
+impl PktNumManager {
+    pub fn new() -> Self {
+        PktNumManager {
+            skip_pn: None,
+            skip_pn_counter: None,
+        }
+    }
+
+    pub fn on_packet_sent(
+        &mut self, cwnd: usize, max_datagram_size: usize,
+        handshake_completed: bool,
+    ) {
+        // Decrement skip_pn_counter for each packet sent
+        if let Some(counter) = &mut self.skip_pn_counter {
+            *counter = counter.saturating_sub(1);
+        } else if self.should_arm_skip_counter(handshake_completed) {
+            self.arm_skip_counter(cwnd, max_datagram_size);
+        }
+    }
+
+    fn should_arm_skip_counter(&self, handshake_completed: bool) -> bool {
+        // Arm if the counter is not set
+        let counter_not_set = self.skip_pn_counter.is_none();
+        // Don't arm until we have verified the current skip_pn. Rearming the
+        // counter could result in overwriting the skip_pn before
+        // validating the current skip_pn.
+        let no_current_skip_packet = self.skip_pn.is_none();
+
+        // Skip pn only after the handshake has completed
+        counter_not_set && no_current_skip_packet && handshake_completed
+    }
+
+    pub fn should_skip_pn(&self, handshake_completed: bool) -> bool {
+        // Only skip a new packet once we have validated  the peer hasn't sent the
+        // current skip packet.  For the purposes of OACK, a skip_pn is
+        // considered validated once we receive an ACK for pn > skip_pn.
+        let no_current_skip_packet = self.skip_pn.is_none();
+        let counter_expired = match self.skip_pn_counter {
+            // Skip if counter has expired
+            Some(counter) => counter == 0,
+            // Don't skip if the counter has not been set
+            None => false,
+        };
+
+        // Skip pn only after the handshake has completed
+        counter_expired && no_current_skip_packet && handshake_completed
+    }
+
+    pub fn skip_pn(&self) -> Option<u64> {
+        self.skip_pn
+    }
+
+    pub fn set_skip_pn(&mut self, skip_pn: Option<u64>) {
+        if skip_pn.is_some() {
+            // Never overwrite skip_pn until the previous one has been verified
+            debug_assert!(self.skip_pn.is_none());
+            // The skip_pn_counter should be expired
+            debug_assert_eq!(self.skip_pn_counter.unwrap(), 0);
+        }
+
+        self.skip_pn = skip_pn;
+        // unset the counter
+        self.skip_pn_counter = None;
+    }
+
+    // Dynamically vary the skip counter based on the CWND.
+    fn arm_skip_counter(&mut self, cwnd: usize, max_datagram_size: usize) {
+        let packets_per_cwnd = (cwnd / max_datagram_size) as u64;
+        let lower = packets_per_cwnd / 2;
+        let upper = packets_per_cwnd * 2;
+        // rand_u64_uniform requires a non-zero value so add 1
+        let skip_range = upper - lower + 1;
+        let rand_skip_value = rand::rand_u64_uniform(skip_range);
+
+        // Skip calculation:
+        // skip_counter = min_skip
+        //                + lower
+        //                + rand(skip_range.lower, skip_range.upper)
+        //
+        //```
+        // c: the current packet number
+        // s: range of random packet number to skip from
+        //
+        // curr_pn
+        //  |
+        //  v               |-lower-|
+        // [c x x x x x x x x x x x s s s s s s s s s s x x]
+        //    |--min_skip---|       |---skip_range----|
+        //
+        //```
+        let skip_pn_counter = MIN_SKIP_COUNTER_VALUE + lower + rand_skip_value;
+
+        self.skip_pn_counter = Some(skip_pn_counter);
+    }
+}
+
 #[derive(Clone, Copy, Default)]
 pub struct PktNumWindow {
     lower: u64,
@@ -980,15 +1162,15 @@ impl PktNumWindow {
     }
 
     fn upper(&self) -> u64 {
-        self.lower
-            .saturating_add(std::mem::size_of::<u128>() as u64 * 8) -
-            1
+        self.lower.saturating_add(size_of::<u128>() as u64 * 8) - 1
     }
 }
 
 #[cfg(test)]
 mod tests {
     use super::*;
+    use crate::test_utils;
+    use crate::MAX_SEND_UDP_PAYLOAD_SIZE;
 
     #[test]
     fn retry() {
@@ -1173,9 +1355,31 @@ mod tests {
     }
 
     #[test]
-    fn pkt_num_decode() {
+    fn pkt_num_encode_decode() {
+        let num_len = pkt_num_len(0, 0);
+        assert_eq!(num_len, 1);
         let pn = decode_pkt_num(0xa82f30ea, 0x9b32, 2);
         assert_eq!(pn, 0xa82f9b32);
+        let mut d = [0; 10];
+        let mut b = octets::OctetsMut::with_slice(&mut d);
+        let num_len = pkt_num_len(0xac5c02, 0xabe8b3);
+        assert_eq!(num_len, 2);
+        encode_pkt_num(0xac5c02, num_len, &mut b).unwrap();
+        // reading
+        let mut b = octets::OctetsMut::with_slice(&mut d);
+        let hdr_num = u64::from(b.get_u16().unwrap());
+        let pn = decode_pkt_num(0xac5c01, hdr_num, num_len);
+        assert_eq!(pn, 0xac5c02);
+        // sending 0xace8fe while having 0xabe8b3 acked
+        let num_len = pkt_num_len(0xace9fe, 0xabe8b3);
+        assert_eq!(num_len, 3);
+        let mut b = octets::OctetsMut::with_slice(&mut d);
+        encode_pkt_num(0xace9fe, num_len, &mut b).unwrap();
+        // reading
+        let mut b = octets::OctetsMut::with_slice(&mut d);
+        let hdr_num = u64::from(b.get_u24().unwrap());
+        let pn = decode_pkt_num(0xace9fa, hdr_num, num_len);
+        assert_eq!(pn, 0xace9fe);
     }
 
     #[test]
@@ -1295,9 +1499,13 @@ mod tests {
 
         let payload_len = b.get_varint().unwrap() as usize;
 
-        let (aead, _) =
-            crypto::derive_initial_key_material(dcid, hdr.version, is_server)
-                .unwrap();
+        let (aead, _) = crypto::derive_initial_key_material(
+            dcid,
+            hdr.version,
+            is_server,
+            false,
+        )
+        .unwrap();
 
         decrypt_hdr(&mut b, &mut hdr, &aead).unwrap();
         assert_eq!(hdr.pkt_num_len, expected_pn_len);
@@ -1510,7 +1718,7 @@ mod tests {
 
         let alg = crypto::Algorithm::ChaCha20_Poly1305;
 
-        let aead = crypto::Open::from_secret(alg, secret.into()).unwrap();
+        let aead = crypto::Open::from_secret(alg, &secret).unwrap();
 
         let mut hdr = Header::from_bytes(&mut b, 0).unwrap();
         assert_eq!(hdr.ty, Type::Short);
@@ -1544,9 +1752,13 @@ mod tests {
 
         b.put_bytes(header).unwrap();
 
-        let (_, aead) =
-            crypto::derive_initial_key_material(dcid, hdr.version, is_server)
-                .unwrap();
+        let (_, aead) = crypto::derive_initial_key_material(
+            dcid,
+            hdr.version,
+            is_server,
+            false,
+        )
+        .unwrap();
 
         let payload_len = frames.len();
 
@@ -1878,7 +2090,7 @@ mod tests {
 
         let alg = crypto::Algorithm::ChaCha20_Poly1305;
 
-        let aead = crypto::Seal::from_secret(alg, secret.into()).unwrap();
+        let aead = crypto::Seal::from_secret(alg, &secret).unwrap();
 
         let pn = 654_360_564;
         let pn_len = 3;
@@ -1930,7 +2142,8 @@ mod tests {
         let payload_len = b.get_varint().unwrap() as usize;
 
         let (aead, _) =
-            crypto::derive_initial_key_material(b"", hdr.version, true).unwrap();
+            crypto::derive_initial_key_material(b"", hdr.version, true, false)
+                .unwrap();
 
         assert_eq!(
             decrypt_pkt(&mut b, 0, 1, payload_len, &aead),
@@ -1963,11 +2176,145 @@ mod tests {
         let payload_len = 1;
 
         let (aead, _) =
-            crypto::derive_initial_key_material(b"", hdr.version, true).unwrap();
+            crypto::derive_initial_key_material(b"", hdr.version, true, false)
+                .unwrap();
 
         assert_eq!(
             decrypt_pkt(&mut b, 0, 1, payload_len, &aead),
             Err(Error::CryptoFail)
         );
     }
+
+    #[test]
+    fn track_largest_packet_sent() {
+        let now = Instant::now();
+        let mut pkt_space = PktNumSpace::new();
+
+        assert!(pkt_space.largest_tx_pkt_num.is_none());
+
+        let sent_ctx = test_utils::helper_packet_sent(1, now, 10);
+        pkt_space.on_packet_sent(&sent_ctx);
+        assert_eq!(pkt_space.largest_tx_pkt_num.unwrap(), 1);
+
+        let sent_ctx = test_utils::helper_packet_sent(2, now, 10);
+        pkt_space.on_packet_sent(&sent_ctx);
+        assert_eq!(pkt_space.largest_tx_pkt_num.unwrap(), 2);
+    }
+
+    #[test]
+    fn skip_pn() {
+        let mut skip_manager = PktNumManager::new();
+        let cwnd = 1000;
+        let handshake_completed = true;
+        let mut next_pn = 0;
+
+        assert!(skip_manager.skip_pn.is_none());
+        assert!(skip_manager.skip_pn_counter.is_none());
+        assert!(!skip_manager.should_skip_pn(handshake_completed));
+
+        // Arm `skip_pn_counter`
+        skip_manager.on_packet_sent(
+            cwnd,
+            MAX_SEND_UDP_PAYLOAD_SIZE,
+            handshake_completed,
+        );
+        assert_eq!(next_pn, 0);
+        assert!(skip_manager.skip_pn.is_none());
+        assert!(skip_manager.skip_pn_counter.unwrap() >= MIN_SKIP_COUNTER_VALUE);
+        assert!(!skip_manager.should_skip_pn(handshake_completed));
+
+        // `should_skip_pn()` should be true once the counter expires
+        while skip_manager.skip_pn_counter.unwrap() > 0 {
+            // pretend to send the next packet
+            next_pn += 1;
+
+            skip_manager.on_packet_sent(
+                cwnd,
+                MAX_SEND_UDP_PAYLOAD_SIZE,
+                handshake_completed,
+            );
+        }
+        assert!(next_pn >= MIN_SKIP_COUNTER_VALUE);
+        assert!(skip_manager.skip_pn.is_none());
+        assert_eq!(skip_manager.skip_pn_counter.unwrap(), 0);
+        assert!(skip_manager.should_skip_pn(handshake_completed));
+
+        // skip the next pkt_num
+        skip_manager.set_skip_pn(Some(next_pn));
+        assert_eq!(skip_manager.skip_pn.unwrap(), next_pn);
+        assert!(skip_manager.skip_pn_counter.is_none());
+        assert!(!skip_manager.should_skip_pn(handshake_completed));
+    }
+
+    #[test]
+    fn arm_skip_counter_only_after_verifying_prev_skip_pn() {
+        let mut skip_manager = PktNumManager::new();
+        let cwnd = 1000;
+        let handshake_completed = true;
+
+        // Set skip pn
+        skip_manager.skip_pn_counter = Some(0);
+        skip_manager.set_skip_pn(Some(42));
+        assert!(skip_manager.skip_pn.is_some());
+        assert!(skip_manager.skip_pn_counter.is_none());
+
+        // Don't arm the skip_pn_counter since its still armed (0 means
+        // expired)
+        skip_manager.on_packet_sent(
+            cwnd,
+            MAX_SEND_UDP_PAYLOAD_SIZE,
+            handshake_completed,
+        );
+        assert!(skip_manager.skip_pn.is_some());
+        assert!(skip_manager.skip_pn_counter.is_none());
+
+        // Arm the skip_pn_counter once the skip_pn has been verified
+        skip_manager.skip_pn = None;
+        skip_manager.on_packet_sent(
+            cwnd,
+            MAX_SEND_UDP_PAYLOAD_SIZE,
+            handshake_completed,
+        );
+        assert!(skip_manager.skip_pn.is_none());
+        assert!(skip_manager.skip_pn_counter.is_some());
+    }
+
+    #[test]
+    fn arm_skip_counter_only_after_handshake_complete() {
+        let mut skip_manager = PktNumManager::new();
+        let cwnd = 1000;
+        skip_manager.skip_pn_counter = None;
+
+        // Don't arm the skip_pn_counter since handshake is not complete
+        let mut handshake_completed = false;
+        skip_manager.on_packet_sent(
+            cwnd,
+            MAX_SEND_UDP_PAYLOAD_SIZE,
+            handshake_completed,
+        );
+        assert!(skip_manager.skip_pn_counter.is_none());
+
+        // Arm counter after handshake complete
+        handshake_completed = true;
+        skip_manager.on_packet_sent(
+            cwnd,
+            MAX_SEND_UDP_PAYLOAD_SIZE,
+            handshake_completed,
+        );
+        assert!(skip_manager.skip_pn_counter.is_some());
+    }
+
+    #[test]
+    fn only_skip_after_handshake_complete() {
+        let mut skip_manager = PktNumManager::new();
+        skip_manager.skip_pn_counter = Some(0);
+
+        let mut handshake_completed = false;
+        // Don't skip since handshake is not complete
+        assert!(!skip_manager.should_skip_pn(handshake_completed));
+
+        handshake_completed = true;
+        // Skip pn after handshake complete
+        assert!(skip_manager.should_skip_pn(handshake_completed));
+    }
 }
