@@ -24,22 +24,29 @@
 // NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 // SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-use std::time;
-
 use std::collections::BTreeMap;
 use std::collections::VecDeque;
+
 use std::net::SocketAddr;
 
+use std::time::Duration;
+use std::time::Instant;
+
 use smallvec::SmallVec;
 
 use slab::Slab;
 
+use crate::Config;
 use crate::Error;
 use crate::Result;
+use crate::StartupExit;
 
 use crate::pmtud;
 use crate::recovery;
+use crate::recovery::Bandwidth;
 use crate::recovery::HandshakeStatus;
+use crate::recovery::OnLossDetectionTimeoutOutcome;
+use crate::recovery::RecoveryOps;
 
 /// The different states of the path validation.
 #[derive(Debug, Copy, Clone, PartialEq, Eq, PartialOrd, Ord)]
@@ -136,12 +143,12 @@ pub struct Path {
     /// Loss recovery and congestion control state.
     pub recovery: recovery::Recovery,
 
-    /// Path MTU discovery state.
-    pub pmtud: pmtud::Pmtud,
+    /// Path MTU discovery state. None if PMTUD is disabled on the path.
+    pub pmtud: Option<pmtud::Pmtud>,
 
     /// Pending challenge data with the size of the packet containing them and
     /// when they were sent.
-    in_flight_challenges: VecDeque<([u8; 8], usize, time::Instant)>,
+    in_flight_challenges: VecDeque<([u8; 8], usize, Instant)>,
 
     /// The maximum challenge size that got acknowledged.
     max_challenge_size: usize,
@@ -150,7 +157,7 @@ pub struct Path {
     probing_lost: usize,
 
     /// Last instant when a probing packet got lost.
-    last_probe_lost_time: Option<time::Instant>,
+    last_probe_lost_time: Option<Instant>,
 
     /// Received challenge data.
     received_challenges: VecDeque<[u8; 8]>,
@@ -167,6 +174,20 @@ pub struct Path {
     /// Total number of packets sent with data retransmitted from this path.
     pub retrans_count: usize,
 
+    /// Total number of times PTO (probe timeout) fired.
+    ///
+    /// Loss usually happens in a burst so the number of packets lost will
+    /// depend on the volume of inflight packets at the time of loss (which
+    /// can be arbitrary). PTO count measures the number of loss events and
+    /// provides a normalized loss metric.
+    pub total_pto_count: usize,
+
+    /// Number of DATAGRAM frames sent on this path.
+    pub dgram_sent_count: usize,
+
+    /// Number of DATAGRAM frames received on this path.
+    pub dgram_recv_count: usize,
+
     /// Total number of sent bytes over this path.
     pub sent_bytes: u64,
 
@@ -207,8 +228,8 @@ impl Path {
     pub fn new(
         local_addr: SocketAddr, peer_addr: SocketAddr,
         recovery_config: &recovery::RecoveryConfig,
-        path_challenge_recv_max_queue_len: usize, pmtud_init: usize,
-        is_initial: bool,
+        path_challenge_recv_max_queue_len: usize, is_initial: bool,
+        config: Option<&Config>,
     ) -> Self {
         let (state, active_scid_seq, active_dcid_seq) = if is_initial {
             (PathState::Validated, Some(0), Some(0))
@@ -216,6 +237,23 @@ impl Path {
             (PathState::Unknown, None, None)
         };
 
+        let pmtud = config.and_then(|c| {
+            if c.pmtud {
+                let maximum_supported_mtu: usize = std::cmp::min(
+                    // if the max_udp_payload_size doesn't fit into a usize, then
+                    // max_send_udp_payload_size must be smaller so use that
+                    c.local_transport_params
+                        .max_udp_payload_size
+                        .try_into()
+                        .unwrap_or(c.max_send_udp_payload_size),
+                    c.max_send_udp_payload_size,
+                );
+                Some(pmtud::Pmtud::new(maximum_supported_mtu))
+            } else {
+                None
+            }
+        });
+
         Self {
             local_addr,
             peer_addr,
@@ -224,7 +262,7 @@ impl Path {
             state,
             active: false,
             recovery: recovery::Recovery::new_with_config(recovery_config),
-            pmtud: pmtud::Pmtud::new(pmtud_init),
+            pmtud,
             in_flight_challenges: VecDeque::new(),
             max_challenge_size: 0,
             probing_lost: 0,
@@ -236,6 +274,9 @@ impl Path {
             sent_count: 0,
             recv_count: 0,
             retrans_count: 0,
+            total_pto_count: 0,
+            dgram_sent_count: 0,
+            dgram_recv_count: 0,
             sent_bytes: 0,
             recv_bytes: 0,
             stream_retrans_bytes: 0,
@@ -336,11 +377,15 @@ impl Path {
         &mut self, hs_confirmed: bool, hs_done: bool, out_len: usize,
         is_closing: bool, frames_empty: bool,
     ) -> bool {
+        let Some(pmtud) = self.pmtud.as_mut() else {
+            return false;
+        };
+
         (hs_confirmed && hs_done) &&
-            self.pmtud.get_probe_size() > self.pmtud.get_current() &&
-            self.recovery.cwnd_available() > self.pmtud.get_probe_size() &&
-            out_len >= self.pmtud.get_probe_size() &&
-            self.pmtud.get_probe_status() &&
+            pmtud.get_probe_size() > pmtud.get_current_mtu() &&
+            self.recovery.cwnd_available() > pmtud.get_probe_size() &&
+            out_len >= pmtud.get_probe_size() &&
+            pmtud.should_probe() &&
             !is_closing &&
             frames_empty
     }
@@ -352,7 +397,7 @@ impl Path {
 
     /// Handles the sending of PATH_CHALLENGE.
     pub fn add_challenge_sent(
-        &mut self, data: [u8; 8], pkt_size: usize, sent_time: time::Instant,
+        &mut self, data: [u8; 8], pkt_size: usize, sent_time: Instant,
     ) {
         self.on_challenge_sent();
         self.in_flight_challenges
@@ -419,10 +464,10 @@ impl Path {
     }
 
     pub fn on_loss_detection_timeout(
-        &mut self, handshake_status: HandshakeStatus, now: time::Instant,
+        &mut self, handshake_status: HandshakeStatus, now: Instant,
         is_server: bool, trace_id: &str,
-    ) -> (usize, usize) {
-        let (lost_packets, lost_bytes) = self.recovery.on_loss_detection_timeout(
+    ) -> OnLossDetectionTimeoutOutcome {
+        let outcome = self.recovery.on_loss_detection_timeout(
             handshake_status,
             now,
             trace_id,
@@ -470,7 +515,16 @@ impl Path {
             }
         }
 
-        (lost_packets, lost_bytes)
+        // Track PTO timeout event
+        self.total_pto_count += 1;
+
+        outcome
+    }
+
+    pub fn reinit_recovery(
+        &mut self, recovery_config: &recovery::RecoveryConfig,
+    ) {
+        self.recovery = recovery::Recovery::new_with_config(recovery_config)
     }
 
     pub fn stats(&self) -> PathStats {
@@ -481,20 +535,33 @@ impl Path {
             active: self.active,
             recv: self.recv_count,
             sent: self.sent_count,
-            lost: self.recovery.lost_count,
+            lost: self.recovery.lost_count(),
             retrans: self.retrans_count,
+            total_pto_count: self.total_pto_count,
+            dgram_recv: self.dgram_recv_count,
+            dgram_sent: self.dgram_sent_count,
             rtt: self.recovery.rtt(),
             min_rtt: self.recovery.min_rtt(),
+            max_rtt: self.recovery.max_rtt(),
             rttvar: self.recovery.rttvar(),
             cwnd: self.recovery.cwnd(),
             sent_bytes: self.sent_bytes,
             recv_bytes: self.recv_bytes,
-            lost_bytes: self.recovery.bytes_lost,
+            lost_bytes: self.recovery.bytes_lost(),
             stream_retrans_bytes: self.stream_retrans_bytes,
             pmtu: self.recovery.max_datagram_size(),
-            delivery_rate: self.recovery.delivery_rate(),
+            delivery_rate: self.recovery.delivery_rate().to_bytes_per_second(),
+            max_bandwidth: self
+                .recovery
+                .max_bandwidth()
+                .map(Bandwidth::to_bytes_per_second),
+            startup_exit: self.recovery.startup_exit(),
         }
     }
+
+    pub fn bytes_in_flight_duration(&self) -> Duration {
+        self.recovery.bytes_in_flight_duration()
+    }
 }
 
 /// An iterator over SocketAddr.
@@ -547,7 +614,6 @@ impl PathMap {
     /// capacity limit.
     pub fn new(
         mut initial_path: Path, max_concurrent_paths: usize, is_server: bool,
-        enable_pmtud: bool, max_send_udp_payload_size: usize,
     ) -> Self {
         let mut paths = Slab::with_capacity(1); // most connections only have one path
         let mut addrs_to_paths = BTreeMap::new();
@@ -558,14 +624,6 @@ impl PathMap {
         // As it is the first path, it is active by default.
         initial_path.active = true;
 
-        // Enable path MTU Discovery and start probing with the largest datagram
-        // size.
-        if enable_pmtud {
-            initial_path.pmtud.should_probe(enable_pmtud);
-            initial_path.pmtud.set_probe_size(max_send_udp_payload_size);
-            initial_path.pmtud.enable(enable_pmtud);
-        }
-
         let active_path_id = paths.insert(initial_path);
         addrs_to_paths.insert((local_addr, peer_addr), active_path_id);
 
@@ -642,13 +700,13 @@ impl PathMap {
 
     /// Returns an iterator over all existing paths.
     #[inline]
-    pub fn iter(&self) -> slab::Iter<Path> {
+    pub fn iter(&self) -> slab::Iter<'_, Path> {
         self.paths.iter()
     }
 
     /// Returns a mutable iterator over all existing paths.
     #[inline]
-    pub fn iter_mut(&mut self) -> slab::IterMut<Path> {
+    pub fn iter_mut(&mut self) -> slab::IterMut<'_, Path> {
         self.paths.iter_mut()
     }
 
@@ -822,13 +880,27 @@ impl PathMap {
 
         Ok(())
     }
+
+    /// Configures path MTU discovery on all existing paths.
+    pub fn set_discover_pmtu_on_existing_paths(
+        &mut self, discover: bool, max_send_udp_payload_size: usize,
+    ) {
+        for (_, path) in self.paths.iter_mut() {
+            path.pmtud = if discover {
+                Some(pmtud::Pmtud::new(max_send_udp_payload_size))
+            } else {
+                None
+            };
+        }
+    }
 }
 
 /// Statistics about the path of a connection.
 ///
-/// It is part of the `Stats` structure returned by the [`stats()`] method.
+/// A connection’s path statistics can be collected using the [`path_stats()`]
+/// method.
 ///
-/// [`stats()`]: struct.Connection.html#method.stats
+/// [`path_stats()`]: struct.Connection.html#method.path_stats
 #[derive(Clone)]
 pub struct PathStats {
     /// The local address of the path.
@@ -855,15 +927,32 @@ pub struct PathStats {
     /// The number of sent QUIC packets with retransmitted data.
     pub retrans: usize,
 
+    /// The number of times PTO (probe timeout) fired.
+    ///
+    /// Loss usually happens in a burst so the number of packets lost will
+    /// depend on the volume of inflight packets at the time of loss (which
+    /// can be arbitrary). PTO count measures the number of loss events and
+    /// provides a normalized loss metric.
+    pub total_pto_count: usize,
+
+    /// The number of DATAGRAM frames received.
+    pub dgram_recv: usize,
+
+    /// The number of DATAGRAM frames sent.
+    pub dgram_sent: usize,
+
     /// The estimated round-trip time of the connection.
-    pub rtt: time::Duration,
+    pub rtt: Duration,
 
     /// The minimum round-trip time observed.
-    pub min_rtt: Option<time::Duration>,
+    pub min_rtt: Option<Duration>,
+
+    /// The maximum round-trip time observed.
+    pub max_rtt: Option<Duration>,
 
     /// The estimated round-trip time variation in samples using a mean
     /// variation.
-    pub rttvar: time::Duration,
+    pub rttvar: Duration,
 
     /// The size of the connection's congestion window in bytes.
     pub cwnd: usize,
@@ -892,6 +981,15 @@ pub struct PathStats {
     /// [`SendInfo.at`]: struct.SendInfo.html#structfield.at
     /// [Pacing]: index.html#pacing
     pub delivery_rate: u64,
+
+    /// The maximum bandwidth estimate for the connection in bytes/s.
+    ///
+    /// Note: not all congestion control algorithms provide this metric;
+    /// it is currently only implemented for bbr2_gcongestion.
+    pub max_bandwidth: Option<u64>,
+
+    /// Statistics from when a CCA first exited the startup phase.
+    pub startup_exit: Option<StartupExit>,
 }
 
 impl std::fmt::Debug for PathStats {
@@ -951,18 +1049,18 @@ mod tests {
             server_addr,
             &recovery_config,
             config.path_challenge_recv_max_queue_len,
-            1200,
             true,
+            None,
         );
-        let mut path_mgr = PathMap::new(path, 2, false, true, 1200);
+        let mut path_mgr = PathMap::new(path, 2, false);
 
         let probed_path = Path::new(
             client_addr_2,
             server_addr,
             &recovery_config,
             config.path_challenge_recv_max_queue_len,
-            1200,
             false,
+            None,
         );
         path_mgr.insert_path(probed_path, false).unwrap();
 
@@ -979,7 +1077,7 @@ mod tests {
         path_mgr.get_mut(pid).unwrap().add_challenge_sent(
             data,
             MIN_CLIENT_INITIAL_LEN - 1,
-            time::Instant::now(),
+            Instant::now(),
         );
 
         assert!(!path_mgr.get_mut(pid).unwrap().validation_requested());
@@ -1009,7 +1107,7 @@ mod tests {
         path_mgr.get_mut(pid).unwrap().add_challenge_sent(
             data,
             MIN_CLIENT_INITIAL_LEN,
-            time::Instant::now(),
+            Instant::now(),
         );
 
         path_mgr.on_response_received(data).unwrap();
@@ -1038,17 +1136,17 @@ mod tests {
             server_addr,
             &recovery_config,
             config.path_challenge_recv_max_queue_len,
-            1200,
             true,
+            None,
         );
-        let mut client_path_mgr = PathMap::new(path, 2, false, false, 1200);
+        let mut client_path_mgr = PathMap::new(path, 2, false);
         let mut server_path = Path::new(
             server_addr,
             client_addr,
             &recovery_config,
             config.path_challenge_recv_max_queue_len,
-            1200,
             false,
+            None,
         );
 
         let client_pid = client_path_mgr
@@ -1061,11 +1159,7 @@ mod tests {
         client_path_mgr
             .get_mut(client_pid)
             .unwrap()
-            .add_challenge_sent(
-                data,
-                MIN_CLIENT_INITIAL_LEN,
-                time::Instant::now(),
-            );
+            .add_challenge_sent(data, MIN_CLIENT_INITIAL_LEN, Instant::now());
 
         // Second probe.
         let data_2 = rand::rand_u64().to_be_bytes();
@@ -1073,11 +1167,7 @@ mod tests {
         client_path_mgr
             .get_mut(client_pid)
             .unwrap()
-            .add_challenge_sent(
-                data_2,
-                MIN_CLIENT_INITIAL_LEN,
-                time::Instant::now(),
-            );
+            .add_challenge_sent(data_2, MIN_CLIENT_INITIAL_LEN, Instant::now());
         assert_eq!(
             client_path_mgr
                 .get(client_pid)
@@ -1130,17 +1220,17 @@ mod tests {
             server_addr,
             &recovery_config,
             config.path_challenge_recv_max_queue_len,
-            1200,
             true,
+            None,
         );
-        let mut client_path_mgr = PathMap::new(path, 2, false, false, 1200);
+        let mut client_path_mgr = PathMap::new(path, 2, false);
         let mut server_path = Path::new(
             server_addr,
             client_addr,
             &recovery_config,
             config.path_challenge_recv_max_queue_len,
-            1200,
             false,
+            None,
         );
 
         let client_pid = client_path_mgr
@@ -1153,11 +1243,7 @@ mod tests {
         client_path_mgr
             .get_mut(client_pid)
             .unwrap()
-            .add_challenge_sent(
-                data,
-                MIN_CLIENT_INITIAL_LEN,
-                time::Instant::now(),
-            );
+            .add_challenge_sent(data, MIN_CLIENT_INITIAL_LEN, Instant::now());
 
         // Second probe.
         let data_2 = rand::rand_u64().to_be_bytes();
@@ -1165,11 +1251,7 @@ mod tests {
         client_path_mgr
             .get_mut(client_pid)
             .unwrap()
-            .add_challenge_sent(
-                data_2,
-                MIN_CLIENT_INITIAL_LEN,
-                time::Instant::now(),
-            );
+            .add_challenge_sent(data_2, MIN_CLIENT_INITIAL_LEN, Instant::now());
         assert_eq!(
             client_path_mgr
                 .get(client_pid)
@@ -1185,11 +1267,7 @@ mod tests {
         client_path_mgr
             .get_mut(client_pid)
             .unwrap()
-            .add_challenge_sent(
-                data_3,
-                MIN_CLIENT_INITIAL_LEN,
-                time::Instant::now(),
-            );
+            .add_challenge_sent(data_3, MIN_CLIENT_INITIAL_LEN, Instant::now());
         assert_eq!(
             client_path_mgr
                 .get(client_pid)
@@ -1205,11 +1283,7 @@ mod tests {
         client_path_mgr
             .get_mut(client_pid)
             .unwrap()
-            .add_challenge_sent(
-                data_4,
-                MIN_CLIENT_INITIAL_LEN,
-                time::Instant::now(),
-            );
+            .add_challenge_sent(data_4, MIN_CLIENT_INITIAL_LEN, Instant::now());
         assert_eq!(
             client_path_mgr
                 .get(client_pid)
